@@ -1,10 +1,8 @@
 "Implements ArrayConstraint"
-from __future__ import unicode_literals
-from .set import ConstraintSet
 from .single_equation import SingleEquationConstraint
 
 
-class ArrayConstraint(SingleEquationConstraint, ConstraintSet):
+class ArrayConstraint(SingleEquationConstraint, list):
     """A ConstraintSet for prettier array-constraint printing.
 
     ArrayConstraint gets its `sub` method from ConstrainSet,
@@ -14,15 +12,17 @@ class ArrayConstraint(SingleEquationConstraint, ConstraintSet):
     be either NomialArrays or Varkeys of VectorVariables.
     """
     def __init__(self, constraints, left, oper, right):
+        self.constraints = constraints
+        list.__init__(self, constraints)
         SingleEquationConstraint.__init__(self, left, oper, right)
-        ConstraintSet.__init__(self, constraints)
 
-    def __nonzero__(self):
-        "Allows the use of '=' NomialArrays as truth elements."
-        if self.oper != "=":
-            return NotImplemented
-        return all(bool(p) for p in self.flat())
+    def __iter__(self):
+        yield from self.constraints.flat
+
+    def lines_without(self, excluded):
+        "Returns lines for indentation in hierarchical printing."
+        return self.str_without(excluded).split("\n")
 
     def __bool__(self):
-        "Allows the use of NomialArrays as truth elements in python3."
-        return self.__nonzero__()
+        "Allows the use of '=' NomialArrays as truth elements."
+        return False if self.oper != "=" else bool(self.constraints.all())
