@@ -1,36 +1,61 @@
 """Defines SolutionArray class"""
-from __future__ import unicode_literals, print_function
 import re
-from collections import Iterable
+import difflib
+from operator import sub
+import warnings as pywarnings
 import pickle
+import gzip
+import pickletools
 import numpy as np
 from .nomials import NomialArray
 from .small_classes import DictOfLists, Strings
-from .small_scripts import mag, isnan, try_str_without
+from .small_scripts import mag, try_str_without
 from .repr_conventions import unitstr, lineagestr
 
 
 CONSTRSPLITPATTERN = re.compile(r"([^*]\*[^*])|( \+ )|( >= )|( <= )|( = )")
 
 VALSTR_REPLACES = [
-    ("+nan", " - "),
+    ("+nan", " nan"),
+    ("-nan", " nan"),
+    ("nan%", "nan "),
     ("nan", " - "),
-    ("-nan", " - "),
-    ("+0 ", " 0 "),
-    ("+0.00 ", " 0.00 "),
-    ("-0.00 ", " 0.00 "),
-    ("+0.0% ", " 0.0  "),
-    ("-0.0% ", " 0.0  ")
 ]
 
 
-def senss_table(data, showvars=(), title="Sensitivities", **kwargs):
+class SolSavingEnvironment:
+    """Temporarily removes construction/solve attributes from constraints.
+
+    This approximately halves the size of the pickled solution.
+    """
+
+    def __init__(self, solarray):
+        self.solarray = solarray
+        self.attrstore = {}
+
+    def __enter__(self):
+        for constraint_attr in ["mfm", "pmap", "bounded", "meq_bounded",
+                                "v_ss", "unsubbed", "varkeys"]:
+            store = {}
+            for constraint in self.solarray["sensitivities"]["constraints"]:
+                if getattr(constraint, constraint_attr, None):
+                    store[constraint] = getattr(constraint, constraint_attr)
+                    delattr(constraint, constraint_attr)
+            self.attrstore[constraint_attr] = store
+
+    def __exit__(self, type_, val, traceback):
+        for constraint_attr, store in self.attrstore.items():
+            for constraint, value in store.items():
+                setattr(constraint, constraint_attr, value)
+
+
+def senss_table(data, showvars=(), title="Variable Sensitivities", **kwargs):
     "Returns sensitivity table lines"
-    if "constants" in data.get("sensitivities", {}):
-        data = data["sensitivities"]["constants"]
+    if "variables" in data.get("sensitivities", {}):
+        data = data["sensitivities"]["variables"]
     if showvars:
         data = {k: data[k] for k in showvars if k in data}
-    return var_table(data, title, sortbyvals=True,
+    return var_table(data, title, sortbyvals=True, skipifempty=True,
                      valfmt="%+-.2g  ", vecfmt="%+-8.2g",
                      printunits=False, minval=1e-3, **kwargs)
 
@@ -38,16 +63,18 @@ def senss_table(data, showvars=(), title="Sensitivities", **kwargs):
 def topsenss_table(data, showvars, nvars=5, **kwargs):
     "Returns top sensitivity table lines"
     data, filtered = topsenss_filter(data, showvars, nvars)
-    title = "Most Sensitive" if not filtered else "Next Largest Sensitivities"
+    title = "Most Sensitive Variables"
+    if filtered:
+        title = "Next Most Sensitive Variables"
     return senss_table(data, title=title, hidebelowminval=True, **kwargs)
 
 
 def topsenss_filter(data, showvars, nvars=5):
     "Filters sensitivities down to top N vars"
-    if "constants" in data.get("sensitivities", {}):
-        data = data["sensitivities"]["constants"]
+    if "variables" in data.get("sensitivities", {}):
+        data = data["sensitivities"]["variables"]
     mean_abs_senss = {k: np.abs(s).mean() for k, s in data.items()
-                      if not isnan(s).any()}
+                      if not np.isnan(s).any()}
     topk = [k for k, _ in sorted(mean_abs_senss.items(), key=lambda l: l[1])]
     filter_already_shown = showvars.intersection(topk)
     for k in filter_already_shown:
@@ -60,53 +87,45 @@ def topsenss_filter(data, showvars, nvars=5):
 def insenss_table(data, _, maxval=0.1, **kwargs):
     "Returns insensitivity table lines"
     if "constants" in data.get("sensitivities", {}):
-        data = data["sensitivities"]["constants"]
+        data = data["sensitivities"]["variables"]
     data = {k: s for k, s in data.items() if np.mean(np.abs(s)) < maxval}
     return senss_table(data, title="Insensitive Fixed Variables", **kwargs)
 
 
 def tight_table(self, _, ntightconstrs=5, tight_senss=1e-2, **kwargs):
     "Return constraint tightness lines"
-    if not self.model:
-        return []
-    title = "Tightest Constraints"
-    try:
-        data = [((-float("%+6.2g" % c.relax_sensitivity), str(c)),
-                 "%+6.2g" % c.relax_sensitivity, id(c), c)
-                for c in self.model.flat()
-                if c.relax_sensitivity >= tight_senss]
-    except AttributeError:
-        print("Constraint %s had no `relax_sensitivity` attribute." % c)
-        return []
-    if not data:
-        lines = ["No constraints had a sensitivity above %+5.1g."
-                 % tight_senss]
+    title = "Most Sensitive Constraints"
+    if len(self) > 1:
+        title += " (in last sweep)"
+        data = sorted(((-float("%+6.2g" % s[-1]), str(c)),
+                       "%+6.2g" % s[-1], id(c), c)
+                      for c, s in self["sensitivities"]["constraints"].items()
+                      if s[-1] >= tight_senss)[:ntightconstrs]
     else:
-        data = sorted(data)[:ntightconstrs]
-        lines = constraint_table(data, **kwargs)
-    lines = [title] + ["-"*len(title)] + lines + [""]
-    if "sweepvariables" in self:
-        lines.insert(1, "(for the last sweep only)")
-    return lines
-
+        data = sorted(((-float("%+6.2g" % s), str(c)), "%+6.2g" % s, id(c), c)
+                      for c, s in self["sensitivities"]["constraints"].items()
+                      if s >= tight_senss)[:ntightconstrs]
+    return constraint_table(data, title, **kwargs)
 
 def loose_table(self, _, min_senss=1e-5, **kwargs):
     "Return constraint tightness lines"
-    if not self.model:
-        return []
-    title = "All Loose Constraints"
-    data = [(0, "", c) for c in self.model.flat()
-            if c.relax_sensitivity <= min_senss]
-    if not data:
-        lines = ["No constraints had a sensitivity below %+6.2g." % min_senss]
+    title = "Insensitive Constraints |below %+g|" % min_senss
+    if len(self) > 1:
+        title += " (in last sweep)"
+        data = [(0, "", id(c), c)
+                for c, s in self["sensitivities"]["constraints"].items()
+                if s[-1] <= min_senss]
     else:
-        lines = constraint_table(data, **kwargs)
-    return [title] + ["-"*len(title)] + lines + [""]
+        data = [(0, "", id(c), c)
+                for c, s in self["sensitivities"]["constraints"].items()
+                if s <= min_senss]
+    return constraint_table(data, title, **kwargs)
 
 
 # pylint: disable=too-many-branches,too-many-locals,too-many-statements
-def constraint_table(data, sortbymodel=True, showmodels=True, **_):
+def constraint_table(data, title, sortbymodel=True, showmodels=True, **_):
     "Creates lines for tables where the right side is a constraint."
+    # TODO: this should support 1D array inputs from sweeps
     excluded = ("units", "unnecessary lineage")
     if not showmodels:
         excluded = ("units", "lineage")  # hide all of it
@@ -151,6 +170,8 @@ def constraint_table(data, sortbymodel=True, showmodels=True, **_):
         constraintlines.append(line)
         lines += [(openingstr + " : ", constraintlines[0])]
         lines += [("", l) for l in constraintlines[1:]]
+    if not lines:
+        lines = [("", "(none)")]
     maxlens = np.max([list(map(len, line)) for line in lines
                       if line[0] != ("modelname",)], axis=0)
     dirs = [">", "<"]  # we'll check lengths before using zip
@@ -158,11 +179,11 @@ def constraint_table(data, sortbymodel=True, showmodels=True, **_):
     fmts = ["{0:%s%s}" % (direc, L) for direc, L in zip(dirs, maxlens)]
     for i, line in enumerate(lines):
         if line[0] == ("modelname",):
-            line = [fmts[0].format(" | "), line[1]]
+            linelist = [fmts[0].format(" | "), line[1]]
         else:
-            line = [fmt.format(s) for fmt, s in zip(fmts, line)]
-        lines[i] = "".join(line).rstrip()
-    return lines
+            linelist = [fmt.format(s) for fmt, s in zip(fmts, line)]
+        lines[i] = "".join(linelist).rstrip()
+    return [title] + ["-"*len(title)] + lines + [""]
 
 
 def warnings_table(self, _, **kwargs):
@@ -172,27 +193,28 @@ def warnings_table(self, _, **kwargs):
     if "warnings" not in self or not self["warnings"]:
         return []
     for wtype in self["warnings"]:
-        lines += [wtype] + ["-"*len(wtype)]
         data_vec = self["warnings"][wtype]
         if not hasattr(data_vec, "shape"):
             data_vec = [data_vec]
         for i, data in enumerate(data_vec):
+            title = wtype
             if len(data_vec) > 1:
-                lines += ["| for sweep %i |" % i]
+                title += " in sweep %i" % i
             if wtype == "Unexpectedly Tight Constraints" and data[0][1]:
                 data = [(-int(1e5*c.relax_sensitivity),
                          "%+6.2g" % c.relax_sensitivity, id(c), c)
                         for _, c in data]
-                lines += constraint_table(data, **kwargs)
+                lines += constraint_table(data, title, **kwargs)
             elif wtype == "Unexpectedly Loose Constraints" and data[0][1]:
                 data = [(-int(1e5*c.rel_diff),
                          "%.4g %s %.4g" % c.tightvalues, id(c), c)
                         for _, c in data]
-                lines += constraint_table(data, **kwargs)
+                lines += constraint_table(data, title, **kwargs)
             else:
+                lines += [title] + ["-"*len(wtype)]
                 for msg, _ in data:
                     lines += [msg, ""]
-            lines += [""]
+                lines += [""]
     return lines
 
 
@@ -204,21 +226,34 @@ TABLEFNS = {"sensitivities": senss_table,
             "warnings": warnings_table,
            }
 
+def unrolled_absmax(values):
+    "From an iterable of numbers and arrays, returns the largest magnitude"
+    finalval, absmaxest = None, 0
+    for val in values:
+        absmaxval = np.abs(val).max()
+        if absmaxval >= absmaxest:
+            absmaxest, finalval = absmaxval, val
+    if getattr(finalval, "shape", None):
+        return finalval[np.unravel_index(np.argmax(np.abs(finalval)),
+                                         finalval.shape)]
+    return finalval
 
-def reldiff(val1, val2):
+
+def cast(function, val1, val2):
     "Relative difference between val1 and val2 (positive if val2 is larger)"
-    if hasattr(val1, "shape") or hasattr(val2, "shape") or val1.magnitude != 0:
-        if hasattr(val1, "shape") and val1.shape:
-            val1_dims = len(val1.shape)
-            if (hasattr(val2, "shape") and val1.shape != val2.shape
-                    and val2.shape[:val1_dims] == val1.shape):
-                val1_ = np.tile(val1.magnitude, val2.shape[val1_dims:]+(1,)).T
-                val1 = val1_ * val1.units
-        # numpy division will warn but return infs
-        return (val2/val1 - 1).to("dimensionless").magnitude
-    elif val2.magnitude == 0:  # both are scalar zeroes
-        return 0
-    return np.inf  # just val1 is a scalar zero
+    with pywarnings.catch_warnings():  # skip those pesky divide-by-zeros
+        pywarnings.simplefilter("ignore")
+        if hasattr(val1, "shape") and hasattr(val2, "shape"):
+            if val1.ndim == val2.ndim:
+                return function(val1, val2)
+            lessdim, dimmest = sorted([val1, val2], key=lambda v: v.ndim)
+            dimdelta = dimmest.ndim - lessdim.ndim
+            add_axes = (slice(None),)*lessdim.ndim + (np.newaxis,)*dimdelta
+            if dimmest is val1:
+                return function(dimmest, lessdim[add_axes])
+            if dimmest is val2:
+                return function(lessdim[add_axes], dimmest)
+        return function(val1, val2)
 
 
 class SolutionArray(DictOfLists):
@@ -252,12 +287,11 @@ class SolutionArray(DictOfLists):
     >>> senss.append(sol["sensitivities"]["variables"]["x_{min}"])
     >>> assert all(np.array(senss) == 1)
     """
-    program = None
-    model = None
+    modelstr = ""
     _name_collision_varkeys = None
-    table_titles = {"sweepvariables": "Sweep Variables",
+    table_titles = {"sweepvariables": "Swept Variables",
                     "freevariables": "Free Variables",
-                    "constants": "Constants",
+                    "constants": "Fixed Variables",  # TODO: change everywhere
                     "variables": "Variables"}
 
     def name_collision_varkeys(self):
@@ -284,171 +318,144 @@ class SolutionArray(DictOfLists):
         posy_subbed = self.subinto(posy)
         return getattr(posy_subbed, "c", posy_subbed)
 
-    def almost_equal(self, sol, reltol=1e-3, sens_abstol=0.01):
+    def almost_equal(self, other, reltol=1e-3, sens_abstol=0.01):
         "Checks for almost-equality between two solutions"
-        selfvars = set(self["variables"])
-        solvars = set(sol["variables"])
-        if selfvars != solvars:
+        svars, ovars = self["variables"], other["variables"]
+        svks, ovks = set(svars), set(ovars)
+        if svks != ovks:
             return False
-        for key in selfvars:
-            if abs(reldiff(self(key), sol(key))) >= reltol:
+        for key in svks:
+            if abs(cast(np.divide, svars[key], ovars[key]) - 1) >= reltol:
                 return False
-            if abs(sol["sensitivities"]["variables"][key]
-                   - self["sensitivities"]["variables"][key]) >= sens_abstol:
+            if abs(self["sensitivities"]["variables"][key]
+                   - other["sensitivities"]["variables"][key]) >= sens_abstol:
                 return False
         return True
 
     # pylint: disable=too-many-locals, too-many-branches, too-many-statements
-    def diff(self, sol, showvars=None, min_percent=1.0,
-             show_sensitivities=True, min_senss_delta=0.1,
-             sortbymodel=True):
+    def diff(self, other, showvars=None, *,
+             constraintsdiff=True, senssdiff=True, sensstol=0.1,
+             absdiff=False, abstol=0, reldiff=True, reltol=1.0, **tableargs):
         """Outputs differences between this solution and another
 
         Arguments
         ---------
-        sol : solution or string
-            Strings are treated as paths to valid pickled solutions
-        min_percent : float
-            The smallest percentage difference in the result to consider
-        show_sensitivities : boolean
-            if True, also computer sensitivity deltas
-        min_senss_delta : float
-            The smallest absolute difference in sensitivities to consider
+        other : solution or string
+            strings will be treated as paths to pickled solutions
+        senssdiff : boolean
+            if True, show sensitivity differences
+        sensstol : float
+            the smallest sensitivity difference worth showing
+        abssdiff : boolean
+            if True, show absolute differences
+        absstol : float
+            the smallest absolute difference worth showing
+        reldiff : boolean
+            if True, show relative differences
+        reltol : float
+            the smallest relative difference worth showing
 
         Returns
         -------
         str
         """
-        if isinstance(sol, Strings):
-            sol = pickle.load(open(sol, "rb"))
-        selfvars = set(self["variables"])
-        solvars = set(sol["variables"])
+        tableargs.update({"hidebelowminval": True, "sortbyvals": True,
+                          "skipifempty": False})
+        if isinstance(other, Strings):
+            if other[-4:] == ".pgz":
+                other = SolutionArray.decompress_file(other)
+            else:
+                other = pickle.load(open(other, "rb"))
+        svars, ovars = self["variables"], other["variables"]
+        lines = ["Solution Diff",
+                 "=============",
+                 "(positive means the argument is smaller)", ""]
+        svks, ovks = set(svars), set(ovars)
         if showvars:
+            lines[0] += " (for selected variables)"
+            lines[1] += "========================="
             showvars = self._parse_showvars(showvars)
-            selfvars = set([k for k in showvars
-                            if k in self["variables"]])
-            solvars = set([k for k in showvars
-                           if k in sol["variables"]])
-        sol_diff = {}
-        for key in selfvars.intersection(solvars):
-            sol_diff[key] = 100*reldiff(self(key), sol(key))
-        lines = var_table(sol_diff, "Solution difference", sortbyvals=True,
-                          valfmt="%+6.1f%%  ", vecfmt="%+6.1f%% ",
-                          printunits=False, minval=min_percent,
-                          sortbymodel=sortbymodel)
-        if showvars:
-            lines[0] += " for variables given in `showvars`"
-            lines[1] += "----------------------------------"
-        if len(lines) > 3:
-            lines.insert(1, "(positive means the argument is bigger)")
-        elif sol_diff:
-            values = []
-            for v in sol_diff.values():
-                if hasattr(v, "shape"):
-                    values.extend(v.flatten().tolist())
-                else:
-                    values.append(v)
-            values = np.array(values)
-            i = np.unravel_index(np.argmax(np.abs(values)), values.shape)
-            lines.insert(2, "The largest difference is %g%%" % values[i])
-
-        if show_sensitivities:
-            senss_delta = {}
-            for key in selfvars.intersection(solvars):
-                if key in sol["sensitivities"]["variables"]:
-                    val1 = self["sensitivities"]["variables"][key]
-                    val2 = sol["sensitivities"]["variables"][key]
-                    if hasattr(val1, "shape") and val1.shape:
-                        val1_dims = len(val1.shape)
-                        if (hasattr(val2, "shape") and val1.shape != val2.shape
-                                and val2.shape[:val1_dims] == val1.shape):
-                            val1 = np.tile(val1,
-                                           val2.shape[val1_dims:]+(1,)).T
-                    senss_delta[key] = val2 - val1
-                elif key in sol["sensitivities"]["variables"]:
-                    print("Key %s is not in this solution's sensitivities"
-                          " but is in those of the argument.")
-                else:  # for variables that just aren't in any constraints
-                    senss_delta[key] = 0
-
-            primal_lines = len(lines)
-            lines += var_table(senss_delta, "Solution sensitivity delta",
-                               sortbyvals=True,
-                               valfmt="%+-6.2f  ", vecfmt="%+-6.2f",
-                               printunits=False, minval=min_senss_delta,
-                               sortbymodel=sortbymodel)
-            if showvars:
-                lines[primal_lines] += " for variables given in `showvars`"
-                lines[primal_lines + 1] += "----------------------------------"
-            if len(lines) > primal_lines + 3:
-                lines.insert(
-                    primal_lines + 1,
-                    "(positive means the argument has a higher sensitivity)")
-            elif senss_delta:
-                absmaxvalue, maxvalue = 0, 0
-                for valarray in senss_delta.values():
-                    if not getattr(valarray, "shape", None):
-                        value = valarray
-                    else:
-                        flatvalarray = valarray.flatten()
-                        value = flatvalarray[np.argmax(np.abs(valarray))]
-                    absvalue = abs(value)
-                    if absvalue > absmaxvalue:
-                        maxvalue = value
-                        absmaxvalue = absvalue
-                lines.insert(
-                    primal_lines + 2,
-                    "The largest sensitivity delta is %+g" % maxvalue)
-
-        if selfvars-solvars:
+            svks = {k for k in showvars if k in svars}
+            ovks = {k for k in showvars if k in ovars}
+        if constraintsdiff and other.modelstr and self.modelstr:
+            if self.modelstr == other.modelstr:
+                lines += ["** no constraint differences **", ""]
+            else:
+                cdiff = ["Constraint Differences",
+                         "**********************"]
+                cdiff.extend(difflib.unified_diff(
+                    self.modelstr.split("\n"), other.modelstr.split("\n"),
+                    fromfile="removed in argument", tofile="added in argument",
+                    lineterm="", n=3))
+                cdiff.insert(4, "")
+                cdiff += ["", "**********************", ""]
+                lines += cdiff
+        if svks - ovks:
             lines.append("Variable(s) of this solution"
                          " which are not in the argument:")
-            lines.append("\n".join("  %s" % key for key in selfvars-solvars))
+            lines.append("\n".join("  %s" % key for key in svks - ovks))
             lines.append("")
-        if solvars-selfvars:
+        if ovks - svks:
             lines.append("Variable(s) of the argument"
                          " which are not in this solution:")
-            lines.append("\n".join("  %s" % key for key in solvars-selfvars))
+            lines.append("\n".join("  %s" % key for key in ovks - svks))
             lines.append("")
+        sharedvks = svks.intersection(ovks)
+        if reldiff:
+            rel_diff = {vk: 100*(cast(np.divide, svars[vk], ovars[vk]) - 1)
+                        for vk in sharedvks}
+            lines += var_table(rel_diff,
+                               "Relative Differences |above %g%%|" % reltol,
+                               valfmt="%+.1f%%  ", vecfmt="%+6.1f%% ",
+                               minval=reltol, printunits=False, **tableargs)
+            if lines[-2][:10] == "-"*10:  # nothing larger than sensstol
+                lines.insert(-1, ("The largest is %+g%%."
+                                  % unrolled_absmax(rel_diff.values())))
+        if absdiff:
+            abs_diff = {vk: cast(sub, svars[vk], ovars[vk]) for vk in sharedvks}
+            lines += var_table(abs_diff,
+                               "Absolute Differences |above %g|" % abstol,
+                               valfmt="%+.2g", vecfmt="%+8.2g",
+                               minval=abstol, **tableargs)
+            if lines[-2][:10] == "-"*10:  # nothing larger than sensstol
+                lines.insert(-1, ("The largest is %+g."
+                                  % unrolled_absmax(abs_diff.values())))
+        if senssdiff:
+            ssenss = self["sensitivities"]["variables"]
+            osenss = other["sensitivities"]["variables"]
+            senss_delta = {vk: cast(sub, ssenss[vk], osenss[vk])
+                           for vk in svks.intersection(ovks)}
+            lines += var_table(senss_delta,
+                               "Sensitivity Differences |above %g|" % sensstol,
+                               valfmt="%+-.2f  ", vecfmt="%+-6.2f",
+                               minval=sensstol, printunits=False, **tableargs)
+            if lines[-2][:10] == "-"*10:  # nothing larger than sensstol
+                lines.insert(-1, ("The largest is %+g."
+                                  % unrolled_absmax(senss_delta.values())))
+        return "\n".join(lines)
 
-        out = "\n".join(lines)
-        return out
-
-    def pickle_prep(self):
-        "After calling this, the SolutionArray is ready to pickle"
-        program, model = self.program, self.model
-        self.program, self.model = None, None
-        cost = self["cost"]
-        self["cost"] = mag(cost)
-        warnings = {}
-        if "warnings" in self:
-            for wtype in self["warnings"]:
-                warnings[wtype] = self["warnings"][wtype]
-                warnarray = np.array(self["warnings"][wtype])
-                warnarray.T[1] = None  # remove pointer to exact constraint
-                if len(warnarray.shape) == 2:
-                    warnarray = warnarray.tolist()
-                self["warnings"][wtype] = warnarray
-        return program, model, cost, warnings
-
-    def save(self, filename="solution.pkl"):
+    def save(self, filename="solution.pkl", **pickleargs):
         """Pickles the solution and saves it to a file.
 
-        The saved solution is identical except for two things:
-            - the cost is made unitless
-            - the solution's 'program' attribute is removed
-            - the solution's 'model' attribute is removed
-            - the data field is removed from the solution's warnings
-                (the "message" field is preserved)
-
         Solution can then be loaded with e.g.:
         >>> import pickle
         >>> pickle.load(open("solution.pkl"))
         """
-        program, model, cost, warnings = self.pickle_prep()
-        pickle.dump(self, open(filename, "wb"), protocol=1)
-        self["cost"], self["warnings"] = cost, warnings
-        self.program, self.model = program, model
+        with SolSavingEnvironment(self):
+            pickle.dump(self, open(filename, "wb"), **pickleargs)
+
+    def save_compressed(self, filename="solution.pgz", **cpickleargs):
+        "Pickle a file and then compress it into a file with extension."
+        with gzip.open(filename, "wb") as f:
+            with SolSavingEnvironment(self):
+                pickled = pickle.dumps(self, **cpickleargs)
+            f.write(pickletools.optimize(pickled))
+
+    @staticmethod
+    def decompress_file(file):
+        "Load a gzip-compressed pickle file"
+        with gzip.open(file, "rb") as f:
+            return pickle.Unpickler(f).load()
 
     def varnames(self, showvars, exclude):
         "Returns list of variables, optionally with minimal unique names"
@@ -457,7 +464,7 @@ class SolutionArray(DictOfLists):
         for key in self.name_collision_varkeys():
             key.descr["necessarylineage"] = True
         names = {}
-        for key in (showvars or self["variables"]):
+        for key in showvars or self["variables"]:
             for k in self["variables"].keymap[key]:
                 names[k.str_without(exclude)] = k
         for key in self.name_collision_varkeys():
@@ -511,12 +518,11 @@ class SolutionArray(DictOfLists):
     def savetxt(self, filename="solution.txt", printmodel=True, **kwargs):
         "Saves solution table as a text file"
         with open(filename, "w") as f:
-            if printmodel and self.model:
-                f.write(str(self.model))
+            if printmodel:
+                f.write(self.modelstr)
             f.write(self.table(**kwargs))
 
-    def savecsv(self, showvars=None, filename="solution.csv", valcols=5,
-                **kwargs):
+    def savecsv(self, showvars=None, filename="solution.csv", valcols=5):
         "Saves primal solution as a CSV sorted by modelname, like the tables."
         data = self["variables"]
         if showvars:
@@ -536,7 +542,7 @@ class SolutionArray(DictOfLists):
             valcols = 1
         if maxspan < valcols:
             valcols = maxspan
-        lines = var_table(data, "", rawlines=True, maxcolumns=valcols, **kwargs)
+        lines = var_table(data, "", rawlines=True, maxcolumns=valcols)
         with open(filename, "w") as f:
             f.write("Model Name,Variable Name,Value(s)" + ","*valcols
                     + "Units,Description\n")
@@ -559,13 +565,15 @@ class SolutionArray(DictOfLists):
         "Returns NomialArray of each solution substituted into posy."
         if posy in self["variables"]:
             return self["variables"](posy)
-        elif not hasattr(posy, "sub"):
+
+        if not hasattr(posy, "sub"):
             raise ValueError("no variable '%s' found in the solution" % posy)
-        elif len(self) > 1:
+
+        if len(self) > 1:
             return NomialArray([self.atindex(i).subinto(posy)
                                 for i in range(len(self))])
-        else:
-            return posy.sub(self["variables"])
+
+        return posy.sub(self["variables"])
 
     def _parse_showvars(self, showvars):
         showvars_out = set()
@@ -633,7 +641,7 @@ class SolutionArray(DictOfLists):
                 cost = self["cost"]  # pylint: disable=unsubscriptable-object
                 if kwargs.get("latex", None):  # cost is not printed for latex
                     continue
-                strs += ["\n%s\n----" % "Cost"]
+                strs += ["\n%s\n------------" % "Optimal Cost"]
                 if len(self) > 1:
                     costs = ["%-8.3g" % c for c in mag(cost[:4])]
                     strs += [" [ %s %s ]" % ("  ".join(costs),
@@ -667,9 +675,6 @@ class SolutionArray(DictOfLists):
             print("SolutionArray.plot only supports 1-dimensional sweeps")
         if not hasattr(posys, "__len__"):
             posys = [posys]
-        for i, posy in enumerate(posys):
-            if posy in [None, "cost"]:
-                posys[i] = self.program[0].cost   # pylint: disable=unsubscriptable-object
         import matplotlib.pyplot as plt
         from .interactive.plot_sweep import assign_axes
         from . import GPBLU
@@ -683,13 +688,12 @@ class SolutionArray(DictOfLists):
         return plt.gcf(), axes
 
 
-# pylint: disable=too-many-statements,too-many-arguments
-# pylint: disable=too-many-branches,too-many-locals
-def var_table(data, title, printunits=True, latex=False, rawlines=False,
+# pylint: disable=too-many-branches,too-many-locals,too-many-statements
+def var_table(data, title, *, printunits=True, latex=False, rawlines=False,
               varfmt="%s : ", valfmt="%-.4g ", vecfmt="%-8.3g",
               minval=0, sortbyvals=False, hidebelowminval=False,
               included_models=None, excluded_models=None, sortbymodel=True,
-              maxcolumns=5, **_):
+              maxcolumns=5, skipifempty=True, **_):
     """
     Pretty string representation of a dict of VarKeys
     Iterable values are handled specially (partial printing)
@@ -721,22 +725,22 @@ def var_table(data, title, printunits=True, latex=False, rawlines=False,
         return []
     decorated, models = [], set()
     for i, (k, v) in enumerate(data.items()):
-        v_arr = np.array([v])
-        notnan = ~isnan(v_arr)
-        if notnan.any() and np.sum(np.abs(v_arr[notnan])) >= minval:
-            if minval and hidebelowminval and len(notnan.shape) > 1:
-                less_than_min = np.abs(v) <= minval
-                v[np.logical_and(~isnan(v), less_than_min)] = 0
-            model = lineagestr(k.lineage) if sortbymodel else ""
-            models.add(model)
-            b = isinstance(v, Iterable) and bool(v.shape)
-            s = k.str_without(("lineage", "vec"))
-            if not sortbyvals:
-                decorated.append((model, b, (varfmt % s), i, k, v))
-            else:  # for consistent sorting, add small offset to negative vals
-                val = np.mean(np.abs(v)) - (1e-9 if np.mean(v) < 0 else 0)
-                sort = (float("%.4g" % -val), k.name)
-                decorated.append((model, sort, b, (varfmt % s), i, k, v))
+        if np.isnan(v).all() or np.nanmax(np.abs(v)) <= minval:
+            continue  # no values below minval
+        if minval and hidebelowminval and getattr(v, "shape", None):
+            v[np.abs(v) <= minval] = np.nan
+        model = lineagestr(k.lineage) if sortbymodel else ""
+        models.add(model)
+        b = bool(getattr(v, "shape", None))
+        s = k.str_without(("lineage", "vec"))
+        if not sortbyvals:
+            decorated.append((model, b, (varfmt % s), i, k, v))
+        else:  # for consistent sorting, add small offset to negative vals
+            val = np.nanmean(np.abs(v)) - (1e-9 if np.nanmean(v) < 0 else 0)
+            sort = (float("%.4g" % -val), k.name)
+            decorated.append((model, sort, b, (varfmt % s), i, k, v))
+    if not decorated and skipifempty:
+        return []
     if included_models:
         included_models = set(included_models)
         included_models.add("")
@@ -762,7 +766,7 @@ def var_table(data, title, printunits=True, latex=False, rawlines=False,
                     lines.append(
                         [r"\multicolumn{3}{l}{\textbf{" + model + r"}} \\"])
             previous_model = model
-        label = var.descr.get('label', '')
+        label = var.descr.get("label", "")
         units = var.unitstr(" [%s] ") if printunits else ""
         if not isvector:
             valstr = valfmt % val
@@ -770,7 +774,7 @@ def var_table(data, title, printunits=True, latex=False, rawlines=False,
             last_dim_index = len(val.shape)-1
             horiz_dim, ncols = last_dim_index, 1  # starting values
             for dim_idx, dim_size in enumerate(val.shape):
-                if dim_size >= ncols and dim_size <= maxcolumns:
+                if ncols <= dim_size <= maxcolumns:
                     horiz_dim, ncols = dim_idx, dim_size
             # align the array with horiz_dim by making it the last one
             dim_order = list(range(last_dim_index))
